/*
   Copyright 2005 Red Hat, Inc.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in
      the documentation and/or other materials provided with the
      distribution.
    * Neither the name of Red Hat, Inc., nor the names of its
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
   IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "../config.h"
#include <sys/types.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <getopt.h>
#include <glob.h>
#include <grp.h>
#include <pwd.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dbus/dbus.h>
#include <libxml/xmlreader.h>
#ifdef SELINUX
#include <selinux/selinux.h>
#include <selinux/context.h>
#include <selinux/flask.h>
#endif
#include "buffer.h"
#include "common.h"
#include "handlers.h"
#include "mainloop.h"
#include "oddjob_dbus.h"
#include "util.h"

#ident "$Id: oddjobd.c,v 1.14 2005/07/20 20:06:33 nalin Exp $"

/* A structure which we use to keep track of outstanding requests. */
struct oddjob_async_task {
	pid_t pid;
	int stdin_fd, stdout_fd, stderr_fd;
	struct oddjob_buffer *stdin_buffer, *stdout_buffer, *stderr_buffer;
	struct oddjob_dbus_message *msg;
	char *path, *interface, *method, **argv;
};

/* A structure which we use to keep track of configured methods. */
static struct {
	struct method {
		/* object path */
		char *object;
		/* interface name */
		char *interface;
		/* method name */
		char *method;
		/* method type -- internal function, or call a helper */
		enum method_type {
			method_internal,
			method_external,
		} type;
		int n_arguments;
		/* function to call, only applies to internal methods */
		oddjob_dbus_handler *handler;
		void *handler_data;
		/* path to exec, whether or not we prepend the caller's name to
		 * the list of arguments, and how we supply arguments, only
		 * applies to external methods */
		char *execpath;
		dbus_bool_t prepend_user;
		enum oddjob_argument_passing_method {
			oddjob_argument_passing_invalid = 0,
			oddjob_argument_passing_stdin = 1,
			oddjob_argument_passing_cmdline = 2,
		} argument_passing_method;
		/* ACLs for the method */
		struct method_acl {
			enum method_acl_sense {
				method_acl_allow,
				method_acl_deny
			} sense;
			dbus_bool_t apply_user;
			char *user;
			dbus_bool_t apply_min_uid;
			unsigned long min_uid;
			dbus_bool_t apply_max_uid;
			unsigned long max_uid;
#ifdef SELINUX
			dbus_bool_t apply_selinux_enforcing;
			dbus_bool_t selinux_enforcing;
			dbus_bool_t apply_selinux_context;
			char *selinux_context;
			dbus_bool_t apply_selinux_user;
			char *selinux_user;
			dbus_bool_t apply_selinux_role;
			char *selinux_role;
			dbus_bool_t apply_selinux_type;
			char *selinux_type;
			dbus_bool_t apply_selinux_range;
			char *selinux_range;
#endif
			struct method_acl *next;
		} *acl;
	} *methods;
	int n_methods;
	/* flags */
	int debug;
	int quit;
#ifdef SELINUX
	dbus_bool_t selinux_enabled, selinux_enforcing;
#endif
} globals = {
	.methods = NULL,
	.n_methods = 0,
	.debug = 0,
	.quit = 0,
#ifdef SELINUX
	.selinux_enabled = FALSE,
	.selinux_enforcing = FALSE,
#endif
};

static void oddjobd_exec_method(struct oddjob_dbus_context *ctx,
				struct oddjob_dbus_message *msg,
				const char *object_path,
				const char *interface_name,
				const char *method_name,
				const char *user,
				unsigned long uid,
				void *data);
static dbus_bool_t load_config(const char *file, dbus_bool_t ignore_missing);

#ifdef SELINUX
/* Check if we have an SELinux match. */
static dbus_bool_t
check_ac_selinux(struct method_acl *acl, const char *selinux_context)
{
	char *ctx;
	const char *user, *role, *type, *range;
	dbus_bool_t ret;
	context_t context;

	if (!acl->apply_selinux_enforcing &&
	    (!globals.selinux_enabled || !globals.selinux_enforcing)) {
		return TRUE;
	}
	if (selinux_context != NULL) {
		context = context_new(selinux_context);
		ctx = context_str(context);
		user = context_user_get(context);
		role = context_role_get(context);
		type = context_type_get(context);
		range = context_range_get(context);
	} else {
		context = NULL;
		ctx = NULL;
		user = NULL;
		role = NULL;
		type = NULL;
		range = NULL;
	}
	ret = ((!acl->apply_selinux_enforcing) ||
	       (!acl->selinux_enforcing == !globals.selinux_enforcing)) &&
	      ((!acl->apply_selinux_context) ||
	       ((ctx != NULL) &&
		(fnmatch(acl->selinux_context, ctx,
		 	 ODDJOB_SECONTEXT_FNMATCH_FLAGS) == 0))) &&
	      ((!acl->apply_selinux_user) ||
	       ((user != NULL) &&
		(fnmatch(acl->selinux_user, user,
		 	 ODDJOB_SEUSER_FNMATCH_FLAGS) == 0))) &&
	      ((!acl->apply_selinux_role) ||
	       ((role != NULL) &&
		(fnmatch(acl->selinux_role, role,
		 	 ODDJOB_SEROLE_FNMATCH_FLAGS) == 0))) &&
	      ((!acl->apply_selinux_type) ||
	       ((type != NULL) &&
		(fnmatch(acl->selinux_type, type,
		 	 ODDJOB_SETYPE_FNMATCH_FLAGS) == 0))) &&
	      ((!acl->apply_selinux_range) ||
	       ((range != NULL) &&
		(fnmatch(acl->selinux_range, range,
		 	 ODDJOB_SERANGE_FNMATCH_FLAGS) == 0)));
	if (context != NULL) {
		context_free(context);
	}
	return ret;
}
#else
/* Ignore SELinux rules. */
static dbus_bool_t
check_ac_selinux(struct method_acl *acl, const char *selinux_context)
{
	return TRUE;
}
#endif

/* Check if a single ACL entry matches the given user and UID. */
static dbus_bool_t
check_ac_match(struct method_acl *acl, const char *user, unsigned long uid,
	       const char *selinux_context)
{
	return ((!acl->apply_user) ||
		(fnmatch(acl->user, user, ODDJOB_USER_FNMATCH_FLAGS) == 0)) &&
	       ((!acl->apply_min_uid) || (uid >= acl->min_uid)) &&
	       ((!acl->apply_max_uid) || (uid <= acl->max_uid)) &&
	       check_ac_selinux(acl, selinux_context);
}

/* Check if there is a match for the given user and UID in the given ACL,
   returning TRUE if there's an access rule which allows the user acces. */
static dbus_bool_t
check_acl(struct method_acl *acl, const char *user, unsigned long uid,
	  const char *selinux_context)
{
	struct method_acl *i;
	for (i = acl; i != NULL; i = i->next) {
		if (i->sense == method_acl_deny) {
			if (check_ac_match(i, user, uid, selinux_context)) {
				if (globals.debug) {
					fprintf(stderr, "Matched deny rule (");
					if (i->apply_user) {
						fprintf(stderr, " user=%s ",
							i->user);
					}
					if (i->apply_min_uid) {
						fprintf(stderr, " min_uid=%lu ",
							i->min_uid);
					}
					if (i->apply_max_uid) {
						fprintf(stderr, " max_uid=%lu ",
							i->max_uid);
					}
#ifdef SELINUX
					if (i->apply_selinux_enforcing) {
						fprintf(stderr, " seenforcing=%s ",
							i->selinux_enforcing ?
							"yes" : "no");
					}
					if (i->apply_selinux_context) {
						fprintf(stderr, " secontext=%s ",
							i->selinux_context);
					}
					if (i->apply_selinux_user) {
						fprintf(stderr, " seuser=%s ",
							i->selinux_user);
					}
					if (i->apply_selinux_role) {
						fprintf(stderr, " serole=%s ",
							i->selinux_role);
					}
					if (i->apply_selinux_type) {
						fprintf(stderr, " setype=%s ",
							i->selinux_type);
					}
					if (i->apply_selinux_range) {
						fprintf(stderr, " serange=%s ",
							i->selinux_range);
					}
#endif
					fprintf(stderr, ")\n");
				}
				return FALSE;
			}
		}
	}
	for (i = acl; i != NULL; i = i->next) {
		if (i->sense == method_acl_allow) {
			if (check_ac_match(i, user, uid, selinux_context)) {
				if (globals.debug) {
					fprintf(stderr, "Matched allow rule (");
					if (i->apply_user) {
						fprintf(stderr, " user=%s ",
							i->user);
					}
					if (i->apply_min_uid) {
						fprintf(stderr, " min_uid=%lu ",
							i->min_uid);
					}
					if (i->apply_max_uid) {
						fprintf(stderr, " max_uid=%lu ",
							i->max_uid);
					}
#ifdef SELINUX
					if (i->apply_selinux_enforcing) {
						fprintf(stderr, " enforcing=%s ",
							i->selinux_enforcing ?
							"yes" : "no");
					}
					if (i->apply_selinux_context) {
						fprintf(stderr, " context=%s ",
							i->selinux_context);
					}
					if (i->apply_selinux_user) {
						fprintf(stderr, " user=%s ",
							i->selinux_user);
					}
					if (i->apply_selinux_role) {
						fprintf(stderr, " role=%s ",
							i->selinux_role);
					}
					if (i->apply_selinux_type) {
						fprintf(stderr, " type=%s ",
							i->selinux_type);
					}
					if (i->apply_selinux_range) {
						fprintf(stderr, " range=%s ",
							i->selinux_range);
					}
#endif
					fprintf(stderr, ")\n");
				}
				return TRUE;
			}
		}
	}
	if (globals.debug) {
		fprintf(stderr, "Fell through ACL, denying access.\n");
	}
	return FALSE;
}

/* Check if an entry in the global methods list matches the supplied
 * object/interface/method. */
static dbus_bool_t
check_method_matches(struct method *candidate,
		     const char *object,
		     const char *interface,
		     const char *method)
{
	if ((fnmatch(candidate->object, object,
		     ODDJOB_OBJECT_FNMATCH_FLAGS) == 0) &&
	    (fnmatch(candidate->interface, interface,
		     ODDJOB_INTERFACE_FNMATCH_FLAGS) == 0) &&
	    (fnmatch(candidate->method, method,
		     ODDJOB_METHOD_FNMATCH_FLAGS) == 0)) {
		return TRUE;
	}
	return FALSE;
}

/* Check if we are in SELinux enforcing mode. */
static void
check_selinux_applicable(void)
{
#ifdef SELINUX
	globals.selinux_enabled = (is_selinux_enabled() != 0);
	if (globals.selinux_enabled) {
		globals.selinux_enforcing = (security_getenforce() != 0);
	} else {
		globals.selinux_enforcing = FALSE;
	}
#endif
}

/* Find the ACL for the specified method and check if the given user/uid is
 * allowed. */
static dbus_bool_t
check_named_acl(const char *object, const char *interface, const char *method,
		const char *user, unsigned long uid,
		const char *selinux_context)
{
	int i;
	check_selinux_applicable();
	for (i = 0; i < globals.n_methods; i++) {
		if (check_method_matches(&globals.methods[i],
					 object, interface, method)) {
			return check_acl(globals.methods[i].acl, user, uid,
					 selinux_context);
		}
	}
	if (globals.debug) {
		fprintf(stderr, "Attempted to check ACL for unknown method, "
			"denying access.\n");
	}
	return FALSE;
}

/* Convenience functions for reading attributes and contents of XML nodes. */
static dbus_bool_t
load_config_xml_node_name_is(xmlNodePtr node, const char *name)
{
	return (node->name != NULL) &&
	       (xmlStrcmp(node->name, (xmlChar*) name) == 0);
}
static dbus_bool_t
load_config_xml_attr_name_is(xmlAttrPtr attr, const char *name)
{
	return (attr->name != NULL) &&
	       (xmlStrcmp(attr->name, (xmlChar*) name) == 0);
}
static const char *
load_config_xml_attr_data(xmlAttrPtr attr)
{
	xmlNodePtr child;
	for (child = attr->children; child != NULL; child = child->next) {
		if ((child->type == XML_TEXT_NODE) &&
		    (child->content != NULL)) {
			return (const char *) child->content;
		}
	}
	return "";
}
static const char *
load_config_xml_node_data(xmlNodePtr node)
{
	xmlNodePtr child;
	for (child = node->children; child != NULL; child = child->next) {
		if ((child->type == XML_TEXT_NODE) &&
		    (child->content != NULL)) {
			return (const char *) child->content;
		}
	}
	return "";
}

/* Load ACLs from the configuration file. */
static dbus_bool_t
load_config_oddjobconfig_oddjob_access(xmlNodePtr cur,
				       const char *object,
				       const char *interface,
				       const char *method,
				       enum method_acl_sense sense)
{
	xmlAttrPtr attr;
	struct method_acl *ac, *acl;
	int i;
	const char *p;
	char *q;

	ac = oddjob_malloc(sizeof(*ac));
	if (ac == NULL) {
		fprintf(stderr, "Out of memory\n");
		return FALSE;
	}
	memset(ac, 0, sizeof(*ac));

	ac->sense = sense;
	if (globals.debug) {
		fprintf(stderr, "Add new %s%s rule ( ",
			sense == method_acl_allow ? "allow" : "",
			sense == method_acl_deny ? "deny" : "");
	}
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "user")) {
			ac->apply_user = TRUE;
			ac->user = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "user=\"%s\" ", ac->user);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "min_uid")) {
			p = load_config_xml_attr_data(attr);
			ac->min_uid = strtoul(p, &q, 0);
			if ((*p != '\0') && (q != NULL) && (*q == '\0')) {
				ac->apply_min_uid = TRUE;
				if (globals.debug) {
					fprintf(stderr, "min_uid=\"%lu\" ",
						ac->min_uid);
				}
			}
		} else
		if (load_config_xml_attr_name_is(attr, "max_uid")) {
			p = load_config_xml_attr_data(attr);
			ac->max_uid = strtoul(p, &q, 0);
			if ((*p != '\0') && (q != NULL) && (*q == '\0')) {
				ac->apply_max_uid = TRUE;
				if (globals.debug) {
					fprintf(stderr, "max_uid=\"%lu\" ",
						ac->max_uid);
				}
			}
#ifdef SELINUX
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_enforcing")) {
			const char *enforcing;
			ac->apply_selinux_enforcing = TRUE;
			ac->selinux_enforcing = TRUE;
			enforcing = load_config_xml_attr_data(attr);
			if (enforcing != NULL) {
				if (strcmp(enforcing, "no") == 0) {
					ac->selinux_enforcing = FALSE;
				} else
				if (strcmp(enforcing, "yes") == 0) {
					ac->selinux_enforcing = TRUE;
				} else {
					fprintf(stderr,
						"Invalid selinux_enforcing "
						"\"%s\" (expected \"yes\" or "
						"\"no\")!\n", enforcing);
					return FALSE;
				}
			}
			if (globals.debug) {
				fprintf(stderr, "selinux_enforcing=\"%s\" ",
					ac->selinux_enforcing ? "yes" : "no");
			}
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_context")) {
			ac->apply_selinux_context = TRUE;
			ac->selinux_context = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "selinux_context=\"%s\" ",
					ac->selinux_context);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_user")) {
			ac->apply_selinux_user = TRUE;
			ac->selinux_user = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "selinux_user=\"%s\" ",
					ac->selinux_user);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_role")) {
			ac->apply_selinux_role = TRUE;
			ac->selinux_role = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "selinux_role=\"%s\" ",
					ac->selinux_role);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_type")) {
			ac->apply_selinux_type = TRUE;
			ac->selinux_type = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "selinux_type=\"%s\" ",
					ac->selinux_type);
			}
		} else
		if (load_config_xml_attr_name_is(attr, "selinux_range")) {
			ac->apply_selinux_range = TRUE;
			ac->selinux_range = oddjob_strdup(load_config_xml_attr_data(attr));
			if (globals.debug) {
				fprintf(stderr, "selinux_range=\"%s\" ",
					ac->selinux_range);
			}
#endif
		}
	}
	if (globals.debug) {
		fprintf(stderr, ") for method \"%s\".\n", method);
	}
	for (i = 0; i < globals.n_methods; i++) {
		if (check_method_matches(&globals.methods[i],
					 object, interface, method)) {
			if (globals.methods[i].acl == NULL) {
				globals.methods[i].acl = ac;
			} else {
				acl = globals.methods[i].acl;
				while (acl->next != NULL) {
					acl = acl->next;
				}
				acl->next = ac;
			}
			return TRUE;
		}
	}
	fprintf(stderr, "Unknown method \"%s\":\"%s\":\"%s\"\n",
		object, interface, method);
	free(ac->user);
	free(ac);
	return FALSE;
}

/* Load the configuration file specified in an <include> node. */
static dbus_bool_t
load_config_oddjobconfig_include(xmlNodePtr cur)
{
	xmlAttrPtr attr;
	dbus_bool_t ignore_missing, okay;
	glob_t globbed;
	unsigned int i;

	ignore_missing = FALSE;
	if (globals.debug) {
		fprintf(stderr, "Include(");
	}
	/* Check if we need to ignore missing or mismatch files. */
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "ignore_missing")) {
			if (globals.debug) {
				fprintf(stderr, "ignore_missing=\"%s\",",
					load_config_xml_attr_data(attr));
			}
			if (strcmp(load_config_xml_attr_data(attr),
				   "yes") == 0) {
				ignore_missing = TRUE;
			}
		}
	}
	if (globals.debug) {
		fprintf(stderr, "\"%s\").\n", load_config_xml_node_data(cur));
	}

	/* Glob for the filename to allow wildcards to work. */
	memset(&globbed, 0, sizeof(globbed));
	if (glob(load_config_xml_node_data(cur), GLOB_NOCHECK,
		 NULL, &globbed) != 0) {
		if (globals.debug) {
			fprintf(stderr, "glob: %s: %s.\n",
				load_config_xml_node_data(cur),
				strerror(errno));
		}
		return FALSE;
	}

	/* Load each one, but only return success if all succeed. */
	for (i = 0, okay = TRUE; i < globbed.gl_pathc; i++) {
		okay = load_config(globbed.gl_pathv[i],
				   ignore_missing) && okay;
	}

	globfree(&globbed);

	return okay;
}

/* Add the specified method to the global method table. */
static void
method_add(const char *object, const char *interface, const char *method,
	   enum method_type type, oddjob_dbus_handler handler,
	   void *handler_data, int n_arguments,
	   dbus_bool_t prepend_user,
	   enum oddjob_argument_passing_method argument_passing_method)
{
	oddjob_resize_array((void**)&globals.methods,
			    sizeof(globals.methods[0]),
			    globals.n_methods,
			    globals.n_methods + 1);
	globals.methods[globals.n_methods].object = oddjob_strdup(object);
	globals.methods[globals.n_methods].interface = oddjob_strdup(interface);
	globals.methods[globals.n_methods].method = oddjob_strdup(method);
	globals.methods[globals.n_methods].type = type;
	globals.methods[globals.n_methods].handler = handler;
	globals.methods[globals.n_methods].handler_data = handler_data;
	globals.methods[globals.n_methods].n_arguments = n_arguments;
	globals.methods[globals.n_methods].prepend_user = prepend_user;
	globals.methods[globals.n_methods].argument_passing_method =
		oddjob_argument_passing_invalid;
	globals.n_methods++;
}

static dbus_bool_t
load_config_oddjobconfig_oddjob_helper(xmlNodePtr cur,
				       const char *object,
				       const char *interface,
				       const char *method)
{
	xmlAttrPtr attr;
	int i, n_arguments;
	const char *execpath, *arguments, *prepend, *arg_method;
	char *p;
	dbus_bool_t prepend_user;
	enum oddjob_argument_passing_method arg_passing_method;
	for (i = 0; i < globals.n_methods; i++) {
		if (check_method_matches(&globals.methods[i],
					 object, interface, method)) {
			break;
		}
	}
	if (i >= globals.n_methods) {
		fprintf(stderr, "internal error %d (no method named \"%s\" "
			"provided as part of interface \"%s\" "
			"on object \"%s\")!",
			__LINE__, method, interface, object);
		return FALSE;
	}
	if (globals.methods[i].type != method_external) {
		fprintf(stderr, "<helper> node for internal method \"%s\" of "
			"interface \"%s\" in object \"%s\" is not allowed!\n",
			method, interface, object);
		return FALSE;
	}
	execpath = NULL;
	arguments = NULL;
	prepend = NULL;
	arg_method = NULL;
	arg_passing_method = oddjob_argument_passing_stdin;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "exec")) {
			execpath = load_config_xml_attr_data(attr);
		}
		if (load_config_xml_attr_name_is(attr, "arguments")) {
			arguments = load_config_xml_attr_data(attr);
		}
		if (load_config_xml_attr_name_is(attr, "prepend_user_name")) {
			prepend = load_config_xml_attr_data(attr);
		}
		if (load_config_xml_attr_name_is(attr,
						 "argument_passing_method")) {
			arg_method = load_config_xml_attr_data(attr);
		}
	}
	if (execpath == NULL) {
		fprintf(stderr, "Required attribute \"exec\" not given!\n");
		return FALSE;
	}
	if (execpath[0] != '/') {
		fprintf(stderr, "\"exec\" value \"%s\" is not "
			"an absolute path!\n", execpath);
		return FALSE;
	}
	n_arguments = 0;
	if (arguments != NULL) {
		n_arguments = strtol(arguments, &p, 0);
		if ((*arguments == '\0') || (p == NULL) || (*p != '\0')) {
			fprintf(stderr, "Invalid argument count \"%s\"!\n",
				arguments);
			return FALSE;
		}
	}
	prepend_user = FALSE;
	if (prepend != NULL) {
		if (strcmp(prepend, "yes") == 0) {
			prepend_user = TRUE;
		} else
		if (strcmp(prepend, "no") == 0) {
			prepend_user = FALSE;
		} else {
			fprintf(stderr, "Invalid prepend_user_name "
				"\"%s\" (expected \"yes\" or \"no\")!\n",
				arg_method);
			return FALSE;
		}
	}
	arg_passing_method = oddjob_argument_passing_stdin;
	if (arg_method != NULL) {
		if (strcmp(arg_method, "stdin") == 0) {
			arg_passing_method = oddjob_argument_passing_stdin;
		} else
		if (strcmp(arg_method, "cmdline") == 0) {
			arg_passing_method = oddjob_argument_passing_cmdline;
		} else {
			fprintf(stderr, "Invalid argument_passing_method "
				"\"%s\" (expected \"stdin\" or \"cmdline\")!\n",
				arg_method);
			return FALSE;
		}
	}
	globals.methods[i].execpath = oddjob_strdup(execpath);
	globals.methods[i].n_arguments = n_arguments;
	globals.methods[i].prepend_user = prepend_user;
	globals.methods[i].argument_passing_method = arg_passing_method;
	if (globals.debug) {
		fprintf(stderr, "Configure external method "
			"\"%s\":\"%s\":\"%s\", exec=\"%s\", "
			"arguments=%d, prepend_user_name=%s, "
			"argument_passing_method=%s.\n",
			object, interface, method, execpath, n_arguments,
			prepend_user ? "yes" : "no",
			(arg_passing_method == oddjob_argument_passing_cmdline) ?
			"cmdline" :
			((arg_passing_method == oddjob_argument_passing_stdin) ?
			 "stdin" : "unknown(bug!)"));
	}
	return TRUE;
}

static dbus_bool_t
load_config_oddjobconfig_method(xmlNodePtr cur,
				const char *object,
				const char *interface)
{
	xmlAttrPtr attr;
	xmlNodePtr child;
	const char *method;
	dbus_bool_t added_here;
	int i;

	method = NULL;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "name")) {
			method = load_config_xml_attr_data(attr);
		}
	}
	if (method == NULL) {
		fprintf(stderr, "<method> node in \"%s\" interface of \"%s\" "
			"object has no \"name\" attribute!\n",
			interface, object);
		return FALSE;
	}
	added_here = FALSE;
	for (i = 0; i < globals.n_methods; i++) {
		if (check_method_matches(&globals.methods[i],
					 object, interface, method)) {
			break;
		}
	}
	if (i >= globals.n_methods) {
		/* check if it's an external method */
		for (child = cur->children;
		     child != NULL;
		     child = child->next) {
			if (load_config_xml_node_name_is(child, "helper")) {
				/* We'd pass the address of the method here,
				 * but the array may be reallocated, so let
				 * main() set it up. */
				method_add(object, interface, method,
					   method_external, oddjobd_exec_method,
					   NULL, 0, FALSE,
					   oddjob_argument_passing_invalid);
				load_config_oddjobconfig_oddjob_helper(child,
								       object,
								       interface,
								       method);
				break;
			}
		}
		added_here = TRUE;
	}
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "helper")) {
			if (added_here == FALSE) {
				fprintf(stderr, "duplicate <helper> node for "
					"\"%s\":\"%s\":\"%s\" method!\n",
					object, interface, method);
				return FALSE;
			}
		}
		if (load_config_xml_node_name_is(child, "deny")) {
			load_config_oddjobconfig_oddjob_access(child,
							       object,
							       interface,
							       method,
							       method_acl_deny);
		}
		if (load_config_xml_node_name_is(child, "allow")) {
			load_config_oddjobconfig_oddjob_access(child,
							       object,
							       interface,
							       method,
							       method_acl_allow);
		}
	}

	return TRUE;
}
static dbus_bool_t
load_config_oddjobconfig_interface(xmlNodePtr cur, const char *object)
{
	xmlAttrPtr attr;
	xmlNodePtr child;
	const char *interface;
	dbus_bool_t parsed = TRUE;

	interface = NULL;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "name")) {
			interface = load_config_xml_attr_data(attr);
		}
	}
	if (interface == NULL) {
		fprintf(stderr, "<interface> node in object \"%s\" has no "
			"\"name\" attribute!\n", object);
		return FALSE;
	}
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "method")) {
			parsed = load_config_oddjobconfig_method(child,
								 object,
								 interface) &&
				 parsed;
		}
	}

	return parsed;
}
static dbus_bool_t
load_config_oddjobconfig_object(xmlNodePtr cur)
{
	xmlAttrPtr attr;
	xmlNodePtr child;
	const char *object;
	dbus_bool_t parsed = TRUE;

	object = NULL;
	for (attr = cur->properties; attr != NULL; attr = attr->next) {
		if (load_config_xml_attr_name_is(attr, "name")) {
			object = load_config_xml_attr_data(attr);
		}
	}
	if (object == NULL) {
		fprintf(stderr, "<object> node has no \"name\" "
			"attribute!\n");
		return FALSE;
	}
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "interface")) {
			parsed = load_config_oddjobconfig_interface(child,
								    object) &&
				 parsed;
		}
	}

	return parsed;
}
static dbus_bool_t
load_config_oddjobconfig(xmlNodePtr cur)
{
	xmlNodePtr child;
	dbus_bool_t parsed;
	parsed = TRUE;
	for (child = cur->children; child != NULL; child = child->next) {
		if (load_config_xml_node_name_is(child, "include")) {
			parsed = load_config_oddjobconfig_include(child) &&
				 parsed;
		}
		if (load_config_xml_node_name_is(child, "object")) {
			parsed = load_config_oddjobconfig_object(child) &&
				 parsed;
		}
	}
	return parsed;
}
static dbus_bool_t
load_config(const char *filename, dbus_bool_t ignore_missing)
{
	xmlDocPtr doc;
	xmlNodePtr cur;
	struct stat st;
	dbus_bool_t parsed;

	if (filename == NULL) {
		filename = SYSCONFDIR "/" PACKAGE "d.conf";
	}

	if (ignore_missing &&
	    (stat(filename, &st) == -1) && (errno == ENOENT)) {
		if (globals.debug) {
			fprintf(stderr, "Ignoring missing file \"%s\".\n",
				filename);
		}
		return TRUE;
	}

	if (globals.debug) {
		fprintf(stderr, "Parsing configuration file \"%s\".\n",
			filename);
	}
	doc = xmlParseFile(filename);
	if (doc == NULL) {
		fprintf(stderr, "Error parsing configuration from \"%s\".\n",
			filename);
		return FALSE;
	}

	parsed = FALSE;
	cur = xmlDocGetRootElement(doc);
	if (cur != NULL) {
		if (load_config_xml_node_name_is(cur, "oddjobconfig")) {
			parsed = load_config_oddjobconfig(cur);
		}
	}

	xmlFreeDoc(doc);

	return parsed;
}
static void
unload_config(void)
{
	struct method_acl *acl, *tofree;
	int i;

	for (i = 0; i < globals.n_methods; i++) {
		oddjob_free(globals.methods[i].object);
		globals.methods[i].object = NULL;
		oddjob_free(globals.methods[i].interface);
		globals.methods[i].interface = NULL;
		oddjob_free(globals.methods[i].method);
		globals.methods[i].method = NULL;
		if (globals.methods[i].execpath) {
			oddjob_free(globals.methods[i].execpath);
			globals.methods[i].execpath = NULL;
		}
		acl = globals.methods[i].acl;
		while (acl != NULL) {
			tofree = acl;
			acl = acl->next;
			oddjob_free(tofree->user);
			oddjob_free(tofree);
		}
		globals.methods[i].acl = NULL;
	}
	oddjob_free(globals.methods);
	globals.methods = NULL;
	globals.n_methods = 0;
}

/* Callback for mainloop waitpid() checks. */
static void
oddjob_pid_service(pid_t pid, int status, void *data)
{
	struct oddjob_async_task *task;
	char outbuf[LINE_MAX];

	task = data;

	/* If the child exited on a signal, return an error. */
	if (WIFSIGNALED(status)) {
		snprintf(outbuf, sizeof(outbuf),
			 "Child exited on signal %d.",
			 WTERMSIG(status));
		if (globals.debug) {
			fprintf(stderr, "%s\n", outbuf);
		}
		oddjob_dbus_send_message_response_error(task->msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							outbuf);
	}

	/* If the child exited normally, even in error, return the result. */
	if (WIFEXITED(status)) {
		if (globals.debug) {
			fprintf(stderr, "Child completed its task "
				"(status=%d, %ld bytes on stdout, "
				"%ld on stderr).\n", WEXITSTATUS(status),
				(long)oddjob_buffer_length(task->stdout_buffer),
				(long)oddjob_buffer_length(task->stderr_buffer));
		}
		oddjob_dbus_send_message_response_success(task->msg,
							  WEXITSTATUS(status),
							  task->stdout_buffer,
							  task->stderr_buffer);
	}
	/* Free the task. */
	task->pid = -1;
	oddjob_buffer_free(task->stdin_buffer);
	oddjob_buffer_free(task->stdout_buffer);
	oddjob_buffer_free(task->stderr_buffer);
	oddjob_dbus_message_free(task->msg);
	oddjob_free(task->path);
	oddjob_free(task->interface);
	oddjob_free(task->method);
	oddjob_free(task->argv);
	oddjob_free(task);
}

/* Handle descriptor-ready status. */
static dbus_bool_t
oddjob_watch_service(int fd, DBusWatchFlags flags, void *data)
{
	struct oddjob_async_task *task;
	dbus_bool_t stop_watching;
	int i, *fdp;
	struct oddjob_buffer *manip_buffer;
	char buf[LINE_MAX];

	task = data;
	stop_watching = FALSE;
	/* Reduce the fd to a buffer and a pointer to its fd. */
	if (fd == task->stdin_fd) {
		fdp = &task->stdin_fd;
		manip_buffer = task->stdin_buffer;
	}
	if ((fd == task->stdout_fd) || (fd == task->stderr_fd)) {
		if (fd == task->stdout_fd) {
			fdp = &task->stdout_fd;
			manip_buffer = task->stdout_buffer;
		} else {
			fdp = &task->stderr_fd;
			manip_buffer = task->stderr_buffer;
		}
	}
	/* Based on the flags, do I/O and buffer management. */
	if (flags & DBUS_WATCH_READABLE) {
		i = read(*fdp, buf, sizeof(buf));
		switch (i) {
		case -1:
		case 0:
			close(*fdp);
			*fdp = -1;
			stop_watching = TRUE;
			break;
		default:
			oddjob_buffer_append(manip_buffer,
					     (const unsigned char *)buf,
					     i);
			break;
		}
	}
	if (flags & DBUS_WATCH_WRITABLE) {
		if (oddjob_buffer_length(manip_buffer) == 0) {
			close(*fdp);
			*fdp = -1;
			stop_watching = TRUE;
		} else {
			i = write(*fdp,
				  oddjob_buffer_data(manip_buffer),
				  oddjob_buffer_length(manip_buffer));
			switch (i) {
			case -1:
				close(*fdp);
				*fdp = -1;
				stop_watching = TRUE;
				break;
			default:
				oddjob_buffer_consume(manip_buffer, i);
				if (oddjob_buffer_length(manip_buffer) == 0) {
					close(*fdp);
					*fdp = -1;
					stop_watching = TRUE;
				}
				break;
			}
		}
	}
	if (flags & DBUS_WATCH_ERROR) {
		close(*fdp);
		*fdp = -1;
		stop_watching = TRUE;
	}
	/* If we're done, then add this task to the PID watch list. */
	if ((task->stdin_fd == -1) &&
	    (task->stdout_fd == -1) &&
	    (task->stderr_fd == -1)) {
		mainloop_pid_add(task->pid, oddjob_pid_service, task);
	}
	return stop_watching;
}

/* Run an external helper. */
static void
oddjobd_exec_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *object_path,
		    const char *interface_name,
		    const char *method_name,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	struct method *method;
	int exec_status[2], child_in[2], child_out[2], child_err[2], null[3], i;
	unsigned int n;
	unsigned char exec_errno, output_char;
	char buf[PATH_MAX];
	int status;
	const char *arg, *helper_name;
	struct oddjob_async_task *task;

	method = data;

	if (globals.debug) {
		fprintf(stderr, "Received request for \"%s\" from %s.\n",
			method->method, user);
	}
	if (!check_acl(method->acl, user, uid,
		       oddjob_dbus_message_get_selinux_context(msg))) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"%s\" denied to %s.\n",
				method->method, user);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}

	/* Open /dev/null three times, just in case stdio aren't already set. */
	if (globals.debug) {
		fprintf(stderr, "Opening /dev/null 3 times.\n");
	}
	for (n = 0; n < sizeof(null) / sizeof(null[0]); n++) {
		null[n] = open("/dev/null", O_RDWR);
		if (null[n] == -1) {
			snprintf(buf, sizeof(buf),
				 "Error at open(/dev/null) %d: %s.",
				 n + 1, strerror(errno));
			if (globals.debug) {
				fprintf(stderr, "%s\n", buf);
			}
			oddjob_dbus_send_message_response_error(msg,
								ODDJOB_INTERFACE_NAME
								".error.internal",
								buf);
			/* Clean up any /dev/null descriptors which we've
			 * already opened. */
			i = n - 1;
			while (i >= 0) {
				close(null[i]);
				i--;
			}
			return;
		}
	}
	/* Create a pipe for returning the exec() failure code. */
	if (pipe(exec_status) == -1) {
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(buf, sizeof(buf),
			 "Error at pipe(1): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	/* Create a trio of pipes for stdio. */
	if (pipe(child_in) == -1) {
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(buf, sizeof(buf),
			 "Error at pipe(2): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	if (pipe(child_out) == -1) {
		close(child_in[0]);
		close(child_in[1]);
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(buf, sizeof(buf),
			 "Error at pipe(3): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	if (pipe(child_err) == -1) {
		close(child_out[0]);
		close(child_out[1]);
		close(child_in[0]);
		close(child_in[1]);
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(buf, sizeof(buf),
			 "Error at pipe(4): %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	if ((fcntl(child_out[0], F_SETFL, O_NONBLOCK) == -1) ||
	    (fcntl(child_err[0], F_SETFL, O_NONBLOCK) == -1)) {
		close(child_err[0]);
		close(child_err[1]);
		close(child_out[0]);
		close(child_out[1]);
		close(child_in[0]);
		close(child_in[1]);
		close(exec_status[0]);
		close(exec_status[1]);
		close(null[0]);
		close(null[1]);
		close(null[2]);
		snprintf(buf, sizeof(buf),
			 "Error at fcntl: %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	/* At this point we *know* that none of pipes are on stdio, so we no
	 * longer need null[]. */
	close(null[0]);
	close(null[1]);
	close(null[2]);
	/* Here's where the heavy lifting starts.  Do all of the memory
	 * allocation which might fail in the child. */
	task = malloc(sizeof(struct oddjob_async_task));
	if (task == NULL) {
		close(child_err[0]);
		close(child_err[1]);
		close(child_out[0]);
		close(child_out[1]);
		close(child_in[0]);
		close(child_in[1]);
		close(exec_status[0]);
		close(exec_status[1]);
		snprintf(buf, sizeof(buf),
			 "Out of memory: %s.", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
	}
	/* Allocate space for a copy of the calling message, the child's
	 * arguments, our environment variables, and I/O buffers. */
	task->msg = oddjob_dbus_message_dup(msg);
	task->stdin_buffer = oddjob_buffer_new(LINE_MAX);
	task->stdout_buffer = oddjob_buffer_new(LINE_MAX);
	task->stderr_buffer = oddjob_buffer_new(LINE_MAX);
	switch (method->argument_passing_method) {
	case oddjob_argument_passing_cmdline:
		n = (3 + oddjob_dbus_message_get_n_args(msg)) * sizeof(char *);
		task->argv = oddjob_malloc0(n);
		n = 0;
		if (strchr(method->execpath, '/') != NULL) {
			task->argv[n++] = strrchr(method->execpath, '/') + 1;
		} else {
			task->argv[n++] = method->execpath;
		}
		if (method->prepend_user) {
			task->argv[n++] = (char *) user;
		}
		for (i = 0; i < method->n_arguments; i++) {
			arg = oddjob_dbus_message_get_arg(msg, i);
			task->argv[n++] = (char *) arg;
		}
		break;
	case oddjob_argument_passing_stdin:
		n = 3 * sizeof(char *);
		task->argv = oddjob_malloc0(n);
		n = 0;
		if (strchr(method->execpath, '/') != NULL) {
			task->argv[n++] = strrchr(method->execpath, '/') + 1;
		} else {
			task->argv[n++] = method->execpath;
		}
		if (method->prepend_user) {
			oddjob_buffer_append(task->stdin_buffer,
					     (const unsigned char *) user, -1);
			oddjob_buffer_append(task->stdin_buffer,
					     (const unsigned char *) "\n", 1);
		}
		for (i = 0; i < method->n_arguments; i++) {
			arg = oddjob_dbus_message_get_arg(msg, i);
			oddjob_buffer_append(task->stdin_buffer,
					     (const unsigned char *) arg, -1);
			oddjob_buffer_append(task->stdin_buffer,
					     (const unsigned char *) "\n", 1);
		}
		break;
	case oddjob_argument_passing_invalid:
		break;
	}
	task->path = oddjob_strdup_printf("%s=%s",
					  ODDJOB_OBJECT_ENV_VAR, object_path);
	task->interface = oddjob_strdup_printf("%s=%s",
					       ODDJOB_INTERFACE_ENV_VAR,
					       interface_name);
	task->method = oddjob_strdup_printf("%s=%s",
					    ODDJOB_METHOD_ENV_VAR,
					    method_name);
	/* Make a note of which descriptors matter. */
	task->stdin_fd = child_in[1];
	task->stdout_fd = child_out[0];
	task->stderr_fd = child_err[0];
	/* All of that done, we can actually start a child. */
	task->pid = fork();
	switch (task->pid) {
	case -1:
		close(exec_status[0]);
		close(exec_status[1]);
		close(child_in[0]);
		close(child_in[1]);
		close(child_out[0]);
		close(child_out[1]);
		close(child_err[0]);
		close(child_err[1]);
		oddjob_buffer_free(task->stdin_buffer);
		oddjob_buffer_free(task->stdout_buffer);
		oddjob_buffer_free(task->stderr_buffer);
		oddjob_dbus_message_free(task->msg);
		oddjob_free(task->argv);
		oddjob_free(task->path);
		oddjob_free(task->interface);
		oddjob_free(task->method);
		oddjob_free(task);
		snprintf(buf, sizeof(buf),
			 "Error at fork(): %s.\n", strerror(errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.internal",
							buf);
		return;
		break;
	case 0:
		if (globals.debug) {
			fprintf(stderr, "Child preparing to exec %s.\n",
				method->execpath);
		}
		/* We're the child -- close the read end of the error pipe, and
		 * mark its write end to be closed on successful exec(). */
		close(exec_status[0]);
		fcntl(exec_status[1], F_SETFD, FD_CLOEXEC);
		/* Set up stdio. */
		close(child_in[1]);
		close(child_out[0]);
		close(child_err[0]);
		dup2(child_in[0], 0);
		dup2(child_out[1], 1);
		dup2(child_err[1], 2);
		/* Close any open descriptors which aren't the exec status pipe
		 * or stdio. */
		for (i = sysconf(_SC_OPEN_MAX) - 1; i > 2; i--) {
			if (i != exec_status[1]) {
				close(i);
			}
		}
		/* set up the environment */
		putenv(task->path);
		putenv(task->interface);
		putenv(task->method);
#ifdef SELINUX
		/* Set up the SELinux execution context. */
		if (globals.selinux_enabled) {
			const char *client_secontext;
			security_context_t client_context, helper_context,
					   exec_context;

			client_secontext = oddjob_dbus_message_get_selinux_context(msg);
			if (client_secontext == NULL) {
				/* Wha....? */
				exec_errno = 0xff;
				write(exec_status[1], &exec_errno, 1);
				_exit(-1);
			}
			if (getfilecon(method->execpath, &helper_context) == -1) {
				/* No label? */
				exec_errno = 0xfe;
				write(exec_status[1], &exec_errno, 1);
				_exit(-1);
			}
			if (security_compute_create(client_secontext,
						    helper_context,
						    SECCLASS_PROCESS,
						    &exec_context) != 0) {
				/* Failed to compute exec context? */
				exec_errno = 0xfd;
				write(exec_status[1], &exec_errno, 1);
				_exit(-1);
			}
			if (setexeccon(exec_context) == -1) {
				/* Failed to set exec context? */
				exec_errno = 0xfc;
				write(exec_status[1], &exec_errno, 1);
				_exit(-1);
			}
		}
#endif
		/* run the helper */
		execv(method->execpath, task->argv);
		/* uh-oh. send errno to the caller and bail */
		exec_errno = errno;
		write(exec_status[1], &exec_errno, 1);
		_exit(-1);
		break;
	default:
		break;
	}
	/* We're the parent -- no need for things which only the child, uh,
	 * needs. */
	oddjob_free(task->argv);
	task->argv = NULL;
	oddjob_free(task->path);
	task->path = NULL;
	oddjob_free(task->interface);
	task->interface = NULL;
	oddjob_free(task->method);
	task->method = NULL;
	/* Check the exec() status, hoping that the child actually gets to run
	 * within a reasonable time. */
	close(exec_status[1]);
	close(child_in[0]);
	close(child_out[1]);
	close(child_err[1]);
	while ((i = (read(exec_status[0], &exec_errno, 1))) == -1) {
		continue;
	}
	close(exec_status[0]);
	/* If we actually got a byte, then it's an exec() errno. */
	if (i == 1) {
		close(task->stdin_fd);
		close(task->stdout_fd);
		close(task->stderr_fd);
		waitpid(task->pid, &status, 0);
		snprintf(buf, sizeof(buf),
			 "Child signalled exec() error: %s.",
			 strerror(exec_errno));
		if (globals.debug) {
			fprintf(stderr, "%s\n", buf);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.exec",
							buf);
		task->pid = getpid();
		oddjob_buffer_free(task->stdin_buffer);
		oddjob_buffer_free(task->stdout_buffer);
		oddjob_buffer_free(task->stderr_buffer);
		oddjob_dbus_message_free(task->msg);
		oddjob_free(task->path);
		oddjob_free(task->interface);
		oddjob_free(task->method);
		oddjob_free(task->argv);
		oddjob_free(task);
		return;
	}
	/* Exec() succeeded -- hook up the buffers to the main loop and wait
	 * for something interesting to happen. */
	mainloop_oddjob_watch_add(task->stdin_fd,
				  DBUS_WATCH_WRITABLE | DBUS_WATCH_ERROR,
				  oddjob_watch_service, task);
	mainloop_oddjob_watch_add(task->stdout_fd,
				  DBUS_WATCH_READABLE | DBUS_WATCH_ERROR,
				  oddjob_watch_service, task);
	mainloop_oddjob_watch_add(task->stderr_fd,
				  DBUS_WATCH_READABLE | DBUS_WATCH_ERROR,
				  oddjob_watch_service, task);
}

/* Get a list of the available methods. */
static void
oddjobd_list_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *object_path,
		    const char *interface_name,
		    const char *method_name,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	int i;
	char *result, *t;
	char fmt[] = "(object=\"%s\",interface=\"%s\",method=\"%s\")";
	if (!check_named_acl(object_path, interface_name, method_name,
			     user, uid,
			     oddjob_dbus_message_get_selinux_context(msg))) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"%s\" denied.\n",
				method_name);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}
	result = NULL;
	for (i = 0; i < globals.n_methods; i++) {
		if (check_acl(globals.methods[i].acl, user, uid,
			      oddjob_dbus_message_get_selinux_context(msg))) {
			if (result == NULL) {
				t = oddjob_malloc0(strlen(fmt) + 2 +
						   strlen(globals.methods[i].object) + 1 +
						   strlen(globals.methods[i].interface) + 1 +
						   strlen(globals.methods[i].method) + 1);
				sprintf(t, fmt,
					globals.methods[i].object,
					globals.methods[i].interface,
					globals.methods[i].method);
				result = t;
			} else {
				t = oddjob_malloc0(strlen(result) + 1 +
						   strlen(fmt) + 2 +
						   strlen(globals.methods[i].object) + 1 +
						   strlen(globals.methods[i].interface) + 1 +
						   strlen(globals.methods[i].method) + 1);
				strcpy(t, result);
				strcat(t, ",");
				sprintf(t + strlen(t), fmt,
					globals.methods[i].object,
					globals.methods[i].interface,
					globals.methods[i].method);
				oddjob_free(result);
				result = t;
			}
		}
	}
	oddjob_dbus_send_message_response_text(msg, 0, result);
	oddjob_free(result);
}

/* Cause the server to shut down. */
static void
oddjobd_quit_method(struct oddjob_dbus_context *ctx,
		    struct oddjob_dbus_message *msg,
		    const char *object_path,
		    const char *interface_name,
		    const char *method_name,
		    const char *user,
		    unsigned long uid,
		    void *data)
{
	if (!check_named_acl(object_path, interface_name, method_name,
			     user, uid,
			     oddjob_dbus_message_get_selinux_context(msg))) {
		if (globals.debug) {
			fprintf(stderr, "Request for \"%s\" denied to %s.\n",
				method_name, user);
		}
		oddjob_dbus_send_message_response_error(msg,
							ODDJOB_INTERFACE_NAME
							".error.acl",
							"ACL does not allow access");
		return;
	}
	if (globals.debug) {
		fprintf(stderr, "Shutting down.\n");
	}
	oddjob_dbus_send_message_response_text(msg, 0, "");
	globals.quit++;
}

int
main(int argc, char **argv)
{
	struct oddjob_dbus_context *ctx;
	FILE *pidfile;
	gid_t gid;
	int c, i;
	const char *basename, *usage;
	struct {
		dbus_bool_t nofork;
		const char *pidfile;
		const char *config;
	} options = {
		.nofork = FALSE,
		.pidfile = NULL,
		.config = NULL,
	};

	/* If we need to provide a usage message, then we need the basename of
	 * the binary. */
	basename = argv[0];
	if (strchr(basename, '/')) {
		basename = strrchr(basename, '/') + 1;
	}

	/* Process command-line options. */
	usage = "Usage: %s [-d] [-n] [-p pidfile] [-c configfile]\n";
	while ((c = getopt(argc, argv, "dnp:c:")) != -1) {
		switch (c) {
		case 'd':
			globals.debug++;
			/* fall through */
		case 'n':
			options.nofork = TRUE;
			break;
		case 'p':
			options.pidfile = optarg;
			break;
		case 'c':
			options.config = optarg;
			break;
		default:
			printf(usage, basename);
			return 1;
			break;
		}
	}
	if ((argc - optind) > 0) {
		printf(usage, basename);
		return 1;
	}

	/* Add our built-in handlers to our method table. */
	method_add(ODDJOB_OBJECT_PATH, ODDJOB_INTERFACE_NAME, "quit",
		   method_internal, oddjobd_quit_method, NULL,
		   0, FALSE, oddjob_argument_passing_invalid);
	method_add(ODDJOB_OBJECT_PATH, ODDJOB_INTERFACE_NAME,
		   ODDJOB_LIST_REQUESTS_METHOD,
		   method_internal, oddjobd_list_method, NULL,
		   0, FALSE, oddjob_argument_passing_invalid);

	/* Read in ACLs and definitions for external handlers. */
	if (!load_config(options.config, FALSE)) {
		fprintf(stderr, "Error loading configuration!\n");
		return 1;
	}

	/* Open a connection to the system message bus. */
	ctx = oddjob_dbus_listener_new(ODDJOB_SERVICE_NAME);
	if (ctx == NULL) {
		fprintf(stderr, "Error connecting to system D-BUS!\n");
		return 2;
	}

	/* Now add our handlers to the list which our D-BUS layer knows. */
	for (i = 0; i < globals.n_methods; i++) {
		/* If the handler used NULL, give it the method table address,
		 * which wasn't fixed until now. */
		oddjob_dbus_listener_add_method(ctx,
						globals.methods[i].object,
						globals.methods[i].interface,
						globals.methods[i].method,
						globals.methods[i].n_arguments,
						globals.methods[i].handler,
						globals.methods[i].handler_data ?
						globals.methods[i].handler_data :
						&globals.methods[i]);
	}

	/* Clear the supplemental group list.  We assume that we were started
	 * with the right UID/GID values, but there should be no need to keep
	 * the rest of the group memberships? */
	gid = getgid();
	if (setgroups(1, &gid) == -1) {
		fprintf(stderr, "Error clearing supplemental group list!\n");
		return 1;
	}

	/* Change the working directory to the root directory. */
	if (chdir("/") == -1) {
		fprintf(stderr, "Error changing working directory to \"/\"!\n");
		return 1;
	}

	/* Create a pidfile. */
	if (options.pidfile != NULL) {
		pidfile = fopen(options.pidfile, "w");
		if (pidfile == NULL) {
			fprintf(stderr, "Error creating pidfile!\n");
			return 1;
		}
		fprintf(pidfile, "%lu", (long) getpid());
		fclose(pidfile);
	}

	/* Become a daemon, changing the working directory to the root,
	 * dissociating from the controlling terminal (if there is one),
	 * calling fork(), and letting the parent exit. */
	if (!options.nofork) {
		if (daemon(0, 0) != 0) {
			fprintf(stderr, "Error becoming a daemon!\n");
			if (options.pidfile != NULL) {
				unlink(options.pidfile);
			}
			return 1;
		}
	}

	/* Now sit! */
	oddjob_dbus_main_init(ctx);
	globals.quit = 0;
	while (globals.quit == 0) {
		oddjob_dbus_main_iterate(ctx);
	}
	oddjob_dbus_main_done(ctx);

	/* Clean up and exit. */
	unload_config();
	oddjob_dbus_listener_free(ctx);
	if (options.pidfile != NULL) {
		unlink(options.pidfile);
	}

	return 0;
}
